from easyecs.cloudformation.template.task_definition import create_task_definition
from easyecs.model.ecs import EcsFileModel


def create_template(
    service_name,
    aws_account_id,
    aws_region,
    vpc_id,
    subnet_ids,
    azs,
    ecs_manifest,
    run=False,
):
    from aws_cdk import App, BootstraplessSynthesizer, Environment, Stack
    from aws_cdk.aws_ec2 import Subnet, SubnetSelection, Vpc

    app = App(outdir="./.cloudformation")
    bootstrapless_synthesizer = BootstraplessSynthesizer()
    stack = Stack(
        app,
        service_name,
        synthesizer=bootstrapless_synthesizer,
        env=Environment(account=aws_account_id, region=aws_region),
    )

    vpc = Vpc.from_vpc_attributes(stack, "vpc", vpc_id=vpc_id, availability_zones=azs)

    subnets = [
        Subnet.from_subnet_id(stack, id=f"container_{i}", subnet_id=subnet_id)
        for i, subnet_id in enumerate(subnet_ids)
    ]
    listener = None
    lb_security_group = None

    subnet_selection = SubnetSelection(subnets=subnets)
    if ecs_manifest.load_balancer:
        listener, lb_security_group = create_load_balancer(stack, ecs_manifest, vpc)
    task_role = create_task_role(stack, service_name, ecs_manifest)
    execution_task_role = create_execution_task_role(stack, service_name, ecs_manifest)
    ecs_cluster = create_ecs_cluster(stack, service_name, vpc)
    log_group = create_log_group(stack, service_name)
    sg = create_security_group(
        stack,
        service_name,
        vpc,
        ecs_manifest,
        lb_security_group,
        ecs_manifest.security_group_id,
    )
    task_definition = create_task_definition(
        stack,
        service_name,
        task_role,
        execution_task_role,
        log_group,
        ecs_manifest,
        run,
    )
    service = create_ecs_service(
        stack, service_name, ecs_cluster, task_definition, subnet_selection, sg
    )
    if ecs_manifest.load_balancer and listener is not None:
        listener.add_targets(
            "NlbTarget",
            port=ecs_manifest.load_balancer.target_group_port,
            targets=[service],
        )

    app.synth()


def create_load_balancer(stack, ecs_manifest: EcsFileModel, vpc):
    from aws_cdk.aws_ec2 import SecurityGroup, Peer, Port, Subnet, SubnetSelection
    from aws_cdk.aws_elasticloadbalancingv2 import NetworkLoadBalancer

    if ecs_manifest.load_balancer:
        if ecs_manifest.load_balancer.security_group_id:
            lb_security_group = SecurityGroup.from_security_group_id(
                stack,
                "nlb_security_group",
                security_group_id=ecs_manifest.load_balancer.security_group_id,
                allow_all_outbound=False,
            )
        else:
            lb_security_group = SecurityGroup(
                stack, "nlb_security_group", vpc=vpc, description="NLB Security Group"
            )
        if ecs_manifest.load_balancer.arn:
            lb = NetworkLoadBalancer.from_network_load_balancer_attributes(
                stack, "nlb", load_balancer_arn=ecs_manifest.load_balancer.arn, vpc=vpc
            )
        else:
            subnets = [
                Subnet.from_subnet_id(stack, f"Subnet{i}", subnet_id)
                for i, subnet_id in enumerate(ecs_manifest.load_balancer.subnets)
            ]
            lb = NetworkLoadBalancer(
                stack,
                "nlb",
                vpc=vpc,
                load_balancer_name=ecs_manifest.load_balancer.load_balancer_name,
                internet_facing=False,
                vpc_subnets=SubnetSelection(subnets=subnets),
                security_groups=[lb_security_group],
            )
        if ecs_manifest.load_balancer.security_group_rules:
            if ecs_manifest.load_balancer.security_group_rules.egress:
                for (
                    egress_rule
                ) in ecs_manifest.load_balancer.security_group_rules.egress:
                    if egress_rule.cidr:
                        lb_security_group.add_egress_rule(
                            peer=(
                                Peer.ipv4(egress_rule.cidr)
                                if egress_rule != "0.0.0.0/0"
                                else Peer.any_ipv4()
                            ),
                            connection=(
                                Port.tcp(egress_rule.port)
                                if egress_rule.port != -1
                                else Port.all_traffic()
                            ),
                            description=egress_rule.name,
                        )
                    elif egress_rule.prefix_list:
                        lb_security_group.add_egress_rule(
                            peer=Peer.prefix_list(egress_rule.prefix_list),
                            connection=(
                                Port.tcp(egress_rule.port)
                                if egress_rule.port != -1
                                else Port.all_traffic()
                            ),
                            description=egress_rule.name,
                        )
                    elif egress_rule.security_group_id:
                        lb_security_group.add_egress_rule(
                            peer=SecurityGroup.from_security_group_id(
                                stack,
                                "egress_rule_sg",
                                security_group_id=egress_rule.security_group_id,
                            ),
                            connection=(
                                Port.tcp(egress_rule.port)
                                if egress_rule.port != -1
                                else Port.all_traffic()
                            ),
                            description=egress_rule.name,
                        )
            if ecs_manifest.load_balancer.security_group_rules.ingress:
                for (
                    ingress_rule
                ) in ecs_manifest.load_balancer.security_group_rules.ingress:
                    if ingress_rule.cidr:
                        lb_security_group.add_ingress_rule(
                            peer=(
                                Peer.ipv4(ingress_rule.cidr)
                                if ingress_rule != "0.0.0.0/0"
                                else Peer.any_ipv4()
                            ),
                            connection=(
                                Port.tcp(ingress_rule.port)
                                if ingress_rule.port != -1
                                else Port.all_traffic()
                            ),
                            description=ingress_rule.name,
                        )
                    elif ingress_rule.prefix_list:
                        lb_security_group.add_ingress_rule(
                            peer=Peer.prefix_list(ingress_rule.prefix_list),
                            connection=(
                                Port.tcp(ingress_rule.port)
                                if ingress_rule.port != -1
                                else Port.all_traffic()
                            ),
                            description=ingress_rule.name,
                        )
                    elif ingress_rule.security_group_id:
                        lb_security_group.add_ingress_rule(
                            peer=SecurityGroup.from_security_group_id(
                                stack,
                                "ingress_rule_sg",
                                security_group_id=ingress_rule.security_group_id,
                            ),
                            connection=(
                                Port.tcp(ingress_rule.port)
                                if ingress_rule.port != -1
                                else Port.all_traffic()
                            ),
                            description=ingress_rule.name,
                        )
        listener = lb.add_listener(
            "NlbListener", port=ecs_manifest.load_balancer.listener_port
        )
        return listener, lb_security_group


def create_task_role(stack, service_name, ecs_manifest):
    from aws_cdk.aws_iam import (
        Effect,
        ManagedPolicy,
        PolicyDocument,
        PolicyStatement,
        ServicePrincipal,
        Role,
    )

    if ecs_manifest.role.arn:
        return Role.from_role_arn(
            scope=stack, id=ecs_manifest.role.arn, role_arn=ecs_manifest.role.arn
        )
    else:
        role_name = f"{service_name}-task-role"
        policy_name = f"{service_name}-task-role-policy"
        assumed_by = ServicePrincipal("ecs-tasks.amazonaws.com")
        managed_policies_raw = ecs_manifest.role.managed_policies
        managed_policies = [
            ManagedPolicy.from_aws_managed_policy_name(managed_policy)
            for managed_policy in managed_policies_raw
        ]
        policy_statements_raw = ecs_manifest.role.statements
        policy_statements = [
            PolicyStatement(
                sid=policy_statement.sid,
                actions=policy_statement.actions,
                resources=policy_statement.resources,
                effect=(
                    Effect.ALLOW if policy_statement.effect == "Allow" else Effect.DENY
                ),
            )
            for policy_statement in policy_statements_raw
        ]
        inline_policies = PolicyDocument(statements=policy_statements)
        role: Role = Role(
            stack,
            role_name,
            assumed_by=assumed_by,
            inline_policies={policy_name: inline_policies},
            managed_policies=managed_policies,
        )
        role.without_policy_updates()
        return role


def create_execution_task_role(stack, service_name, ecs_manifest):
    from aws_cdk.aws_iam import (
        Effect,
        ManagedPolicy,
        PolicyDocument,
        PolicyStatement,
        ServicePrincipal,
        Role,
    )

    if ecs_manifest.execution_role.arn:
        return Role.from_role_arn(
            scope=stack,
            id=ecs_manifest.execution_role.arn,
            role_arn=ecs_manifest.execution_role.arn,
        )
    else:
        role_name = f"{service_name}-execution-task-role"
        policy_name = f"{service_name}-execution-task-role-policy"
        assumed_by = ServicePrincipal("ecs-tasks.amazonaws.com")
        managed_policies_raw = ecs_manifest.execution_role.managed_policies
        managed_policies = [
            ManagedPolicy.from_aws_managed_policy_name(managed_policy)
            for managed_policy in managed_policies_raw
        ]
        policy_statements_raw = ecs_manifest.execution_role.statements
        policy_statements = [
            PolicyStatement(
                sid=policy_statement.sid,
                actions=policy_statement.actions,
                resources=policy_statement.resources,
                effect=(
                    Effect.ALLOW if policy_statement.effect == "Allow" else Effect.DENY
                ),
            )
            for policy_statement in policy_statements_raw
        ]
        inline_policies = PolicyDocument(statements=policy_statements)
        role: Role = Role(
            stack,
            role_name,
            assumed_by=assumed_by,
            inline_policies={policy_name: inline_policies},
            managed_policies=managed_policies,
        )
        role.without_policy_updates()
        return role


def create_ecs_cluster(stack, service_name, vpc):
    from aws_cdk.aws_ecs import Cluster

    cluster_name = f"{service_name}-cluster"
    return Cluster(stack, cluster_name, cluster_name=cluster_name, vpc=vpc)


def create_log_group(stack, service_name):
    from aws_cdk.aws_logs import LogGroup

    log_group_name = f"{service_name}-log"
    return LogGroup(stack, log_group_name)


def create_security_group(
    stack,
    service_name,
    vpc,
    ecs_manifest,
    lb_security_group,
    security_group_id: str = None,
):
    from aws_cdk.aws_ec2 import ISecurityGroup, SecurityGroup, Port

    sg_name = f"{service_name}-sg"
    if security_group_id:
        sg = SecurityGroup.from_security_group_id(
            stack, sg_name, security_group_id=security_group_id
        )
    else:
        sg: ISecurityGroup = SecurityGroup(
            stack, sg_name, vpc=vpc, allow_all_outbound=True
        )
    if ecs_manifest.load_balancer:
        sg.add_ingress_rule(
            peer=lb_security_group,
            connection=Port.tcp(ecs_manifest.load_balancer.target_group_port),
            description="Allow Port from Load Balancer",
        )
    return sg


def create_ecs_service(
    stack, service_name, cluster, task_definition, subnets, security_group
):
    from aws_cdk.aws_ecs import FargateService

    service_name = f"{service_name}-service"
    return FargateService(
        stack,
        service_name,
        service_name=service_name,
        cluster=cluster,
        task_definition=task_definition,
        vpc_subnets=subnets,
        security_groups=[security_group],
        enable_execute_command=True,
        min_healthy_percent=0,
    )



import json
import os
import random
import re
from typing import Dict
import boto3
import socket
from botocore.exceptions import UnauthorizedSSOTokenError

from easyecs.helpers.color import Color


def load_template(stack_name: str) -> Dict:
    """
    Loads the CloudFormation template from a JSON file.
    """
    with open(f".cloudformation/{stack_name}.template.json") as f:
        return json.load(f)


def template_with_env_var(value: str) -> str:
    template_pattern = r"{{.[^}]+}}"
    tpl_values = re.findall(template_pattern, value)
    for tpl_value in tpl_values:
        env_var_name = tpl_value[3:][:-2]
        new_value = os.environ.get(env_var_name)
        if new_value is None:
            print(
                f"{Color.RED}Environment variable {env_var_name} does not"
                f" exist{Color.END}"
            )
        value = value.replace(tpl_value, new_value)
    return value


def parse_dict_with_env_var(build_args):
    for key, value in build_args.items():
        build_args[key] = template_with_env_var(value)
    return build_args


def convert_tags_to_dict(tags):
    return {tag["Key"]: tag["Value"] for tag in tags}


def convert_containers_to_dict(containers):
    parsed_containers = {}
    for container in containers:
        container_name = container["name"]
        # We give him a target id so it's easier to use it with ssm start session
        runtime_id = container["runtimeId"]
        task_id = runtime_id.split("-")[0]
        target_id = f"{task_id}_{runtime_id}"
        container["targetId"] = target_id
        parsed_containers[container_name] = container
    return parsed_containers


def is_port_in_use(port: int) -> bool:
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.settimeout(0.5)
    try:
        s.connect(("127.0.0.1", int(port)))
        s.shutdown(2)
        return True
    except Exception:
        return False


def check_credentials() -> None:
    client = boto3.client("s3")
    try:
        client.list_buckets()
    except UnauthorizedSSOTokenError:
        print(
            f"{Color.RED}You should refresh your credentials or connect to your aws"
            f" account (aws sso login){Color.END}"
        )
        exit(-1)


def generate_random_port():
    random_port = random.randint(1024, 65535)
    while is_port_in_use(random_port):
        random_port = random.randint(1024, 65535)
    random_port = str(random_port)
    return random_port



import boto3
from botocore.utils import ClientError
from easyecs.cloudformation.client import get_client_cloudformation
from easyecs.cloudformation.fetch import fetch_stack_url
from easyecs.cloudformation.stack.waiter import (
    wait_for_stack_create,
    wait_for_stack_delete,
    wait_for_stack_rollback,
)
from easyecs.helpers.color import Color

from easyecs.helpers.loader import Loader


def handle_delete_error(e: ClientError, stack_name: str, loader: Loader):
    """
    Handles a CloudFormation stack update failure.
    Depending on the error, either waits for a rollback, cancels an update,
    or prints the error and breaks the loop.
    """
    message = e.response["Error"]["Message"]
    if "UPDATE_IN_PROGRESS" in message:
        cloudformation = boto3.resource("cloudformation")
        stack = cloudformation.Stack(stack_name)
        stack.cancel_update()
        wait_for_stack_rollback(stack_name)
        delete_cloudformation_stack(stack_name)
        loader.stop()
    elif "CREATE_FAILED" in message:
        print(f"{Color.RED}Creation failed, please check CloudFormation.{Color.END}")
        loader.stop()
    elif "ROLLBACK_FAILED" in message:
        print(f"{Color.RED}Rollback failed, please check CloudFormation.{Color.END}")
        loader.stop()
    elif "ROLLBACK_IN_PROGRESS" in message:
        wait_for_stack_rollback(stack_name)
        loader.stop()
    elif "CREATE_IN_PROGRESS" in message:
        wait_for_stack_create(stack_name)
        loader.stop()
    else:
        loader.stop_error()
        print(e)


def delete_cloudformation_stack(stack_name: str):
    client = get_client_cloudformation()
    client.delete_stack(StackName=stack_name)


def delete_stack(stack_name):
    """
    Deletes a CloudFormation stack with a given stack_name
    """
    # Initialize the loader with appropriate messages
    loader = Loader(
        "Deleting CloudFormation stack:",
        "Deleting CloudFormation stack: \u2705",
        "Deleting CloudFormation stack: \u274c",
        0.05,
    )
    loader.start()

    try:
        loader.set_metadata(f"Cloudformation URL: {fetch_stack_url(stack_name)}")
        delete_cloudformation_stack(stack_name)
        wait_for_stack_delete(stack_name)
        loader.stop()
    except ClientError as e:
        handle_delete_error(e, stack_name, loader)



import boto3
from easyecs.helpers.common import (
    convert_containers_to_dict,
    convert_tags_to_dict,
)

from easyecs.helpers.selector import select_action


def fetch_aws_account():
    aws_account = boto3.client("iam").list_account_aliases()["AccountAliases"][0]
    return aws_account


def fetch_region():
    client = boto3.client("ec2")
    response = client.describe_regions()
    regions = response["Regions"]
    parsed_regions = [region["RegionName"] for region in regions]
    selected_region: str = select_action(parsed_regions, "Select your ECS region")
    return selected_region


def fetch_vpc_id():
    client = boto3.client("ec2")
    response = client.describe_vpcs()
    vpcs = response["Vpcs"]
    vpcs_parsed = {}
    for vpc in vpcs:
        vpc_id = vpc["VpcId"]
        tags = convert_tags_to_dict(vpc.get("Tags", {}))
        vpc_name = tags.get("Name", "Unnamed VPC")
        vpc_key = f"{vpc_id} - {vpc_name}"
        vpcs_parsed[vpc_key] = vpc
    selected_vpc = select_action(vpcs_parsed.keys(), "Select your ECS Task VPC")
    if not selected_vpc:
        raise Exception("You should selected one VPC!")
    return vpcs_parsed[selected_vpc]["VpcId"]


def fetch_container_subnet_ids(vpc_id):
    client = boto3.client("ec2")
    response = client.describe_subnets(
        Filters=[
            {
                "Name": "vpc-id",
                "Values": [
                    vpc_id,
                ],
            }
        ],
    )
    subnets = response["Subnets"]
    subnets_parsed = {}
    for subnet in subnets:
        subnet_id = subnet["SubnetId"]
        tags = convert_tags_to_dict(subnet.get("Tags", {}))
        subnet_name = tags.get("Name", "Unnamed Subnet")
        subnet_key = f"{subnet_id} - {subnet_name}"
        subnets_parsed[subnet_key] = subnet
    selected_subnets = select_action(
        subnets_parsed.keys(), "Select your ECS Task Subnet(s)", multi=True
    )
    if not selected_subnets:
        raise Exception("You should selected at least one subnet!")
    return [
        subnets_parsed[selected_subnet]["SubnetId"]
        for selected_subnet in selected_subnets
    ]


def fetch_availability_zones(aws_region):
    client = boto3.client("ec2")
    response = client.describe_availability_zones(
        Filters=[{"Name": "region-name", "Values": [aws_region]}]
    )
    azs = response["AvailabilityZones"]
    parsed_azs = [az["ZoneName"] for az in azs]
    return parsed_azs


def fetch_account_id():
    client = boto3.client("sts")
    account_id = client.get_caller_identity()["Account"]
    return account_id


def fetch_is_stack_created(stack_name):
    client = boto3.client("cloudformation")
    try:
        client.describe_stacks(StackName=stack_name)
    except client.exceptions.ClientError:
        return False
    return True


def fetch_containers(user, app_name):
    cluster_name = f"{user}-{app_name}-cluster"
    client = boto3.client("ecs")
    res = client.list_tasks(cluster=cluster_name)
    task_arns = res["taskArns"]
    res_task = client.describe_tasks(cluster=cluster_name, tasks=task_arns)
    containers = res_task["tasks"][0]["containers"]
    # This inject the target for easier use with SSM.
    for container in containers:
        runtime_id = container["runtimeId"]
        task_id = runtime_id.split("-")[0]
        target = f"ecs:{cluster_name}_{task_id}_{runtime_id}"
        container["ssm_target"] = target
    parsed_containers = convert_containers_to_dict(containers)
    return parsed_containers


def fetch_session_region():
    my_session = boto3.session.Session()
    region_name = my_session.region_name
    return region_name


def fetch_stack_url(stack_name):
    client = boto3.client("cloudformation")
    res = client.describe_stacks(StackName=stack_name)
    stack_arn = res["Stacks"][0]["StackId"]
    region_name = fetch_session_region()
    url = f"https://{region_name}.console.aws.amazon.com/cloudformation/home?region={region_name}#/stacks/stackinfo?stackId={stack_arn}"  # noqa: E501
    return url


def fetch_load_balancer_dns(stack_name):
    client_cf = boto3.client("cloudformation")
    res = client_cf.describe_stack_resources(StackName=stack_name)
    resources = res["StackResources"]
    load_balancer = list(
        filter(
            lambda x: x["ResourceType"] == "AWS::ElasticLoadBalancingV2::LoadBalancer",
            resources,
        )
    )
    if len(load_balancer) > 0:
        load_balancer_arn = load_balancer[0]["PhysicalResourceId"]
        client_elbv2 = boto3.client("elbv2")
        load_balancer_dns = client_elbv2.describe_load_balancers(
            LoadBalancerArns=[load_balancer_arn]
        )["LoadBalancers"][0]["DNSName"]
        return load_balancer_dns
    else:
        return None



import os
import re
from typing import Any, Dict, List, Optional, Union
from pydantic import BaseModel, computed_field, field_validator, model_validator
from pathlib import Path

from easyecs.helpers.exceptions import FileNotFoundException


class EcsFileMetadataModel(BaseModel):
    appname: str
    user: str = os.environ["USER"]


class EcsFileStatementModel(BaseModel):
    sid: str
    resources: List[str]
    actions: List[str]
    effect: str


class EcsFileRoleModel(BaseModel):
    arn: Optional[str] = None
    managed_policies: List[str] = []
    statements: List[EcsFileStatementModel] = []

    @field_validator("arn")
    def set_arn(cls, arn):
        arn_pattern: str = r"^arn:aws:iam::\d{0,12}:role\/[\w\d_\/.-]*$"
        assert re.match(
            arn_pattern, arn
        ), f"Role ARN does not respect arn pattern : {arn}"
        return arn

    @field_validator("statements")
    def validate_unique_sid(cls, statements):
        seen = set()
        for statement in statements:
            if statement.sid in seen:
                raise ValueError(f"Duplicate sid {statement.sid} found in statements.")
            seen.add(statement.sid)
        return statements


class EcsFileLimitsModel(BaseModel):
    cpu: int
    memory: int


class EcsFileResourcesModel(BaseModel):
    limits: EcsFileLimitsModel


class EcsFileBuildModel(BaseModel):
    dockerfile: str = "Dockerfile"
    target: Optional[str] = None
    args: Dict[str, str] = {}
    context: str = "."


class EcsFileEnvModel(BaseModel):
    name: str
    value: str
    active: bool = True


class EcsFileSecretModel(BaseModel):
    name: str
    arn: str
    field: str
    active: bool = True


class EcsFileSecretModelV2(BaseModel):
    name: str
    valueFrom: str


class EcsFileVolumeModel(BaseModel):
    name: str
    id: str
    mount_point: str

    @field_validator("id")
    def set_id(cls, id):
        assert id.startswith("fs-"), f"EFS ID should start with fs- instead got {id}"
        return id


class EcsFileContainerHealthCheckModel(BaseModel):
    command: Union[List[str], str]
    interval: int
    retries: int
    start_period: int
    timeout: int


class EcsFileContainerModel(BaseModel):
    name: str
    image: str
    user: str = "root"
    tty: bool = False
    essential: bool = True
    entry_point: Union[Optional[str], Optional[List[str]]] = None
    command: Union[Optional[str], Optional[List[str]]] = None
    resources: EcsFileResourcesModel
    build: Optional[EcsFileBuildModel] = None
    port_forward: List[str] = []
    env: List[EcsFileEnvModel] | Dict[str, Any] | None = None
    secrets: List[EcsFileSecretModel] | List[EcsFileSecretModelV2] = []
    efs_volumes: List[EcsFileVolumeModel] = []
    volumes: List[str] = []
    volumes_excludes: List[str] = []
    healthcheck: Optional[EcsFileContainerHealthCheckModel] = None
    depends_on: Optional[Dict[str, Dict[str, str]]] = None
    ports: Optional[List[str]] = []

    @field_validator("volumes")
    def validate_volumes(cls, volumes):
        resolved_volumes = []
        for volume in volumes:
            _from, _to = volume.split(":")
            if not os.path.exists(_from):
                raise FileNotFoundException(f"{_from} does not exists!")
            resolved_from_dir = Path(_from).parent.resolve()
            resolved_from_file = Path(_from).name
            resolved_volumes.append(f"{resolved_from_dir}/{resolved_from_file}:{_to}")
        return resolved_volumes

    @field_validator("volumes_excludes")
    def validate_volumes_excludes(cls, volumes):
        resolved_volumes = []
        for volume in volumes:
            if not os.path.exists(volume):
                raise FileNotFoundException(f"{volume} does not exists!")
            resolved_from_dir = Path(volume).parent.resolve()
            resolved_from_file = Path(volume).name
            resolved_volumes.append(f"{resolved_from_dir}/{resolved_from_file}")
        return resolved_volumes


class EcsTaskDefinitionModel(BaseModel):
    resources: EcsFileResourcesModel
    containers: List[EcsFileContainerModel]

    @field_validator("containers")
    def validate_single_tty(cls, containers):
        tty_containers = [container.name for container in containers if container.tty]
        if len(tty_containers) > 1:
            raise ValueError(
                "More than one container has tty set to true:"
                f" {', '.join(tty_containers)}."
            )
        return containers

    @computed_field(return_type=List[EcsFileVolumeModel])
    @property
    def efs_volumes(self) -> List[EcsFileVolumeModel]:
        efs_volumes = [
            volume for container in self.containers for volume in container.efs_volumes
        ]
        return efs_volumes


class SecurityGroupRule(BaseModel):
    name: str
    port: int
    cidr: Optional[str] = None
    prefix_list: Optional[str] = None
    security_group_id: Optional[str] = None

    @model_validator(mode="after")
    def validate_cidr(self):
        has_at_least_two_true = lambda lst: sum(lst) >= 2  # noqa: E731
        if has_at_least_two_true(
            [
                self.prefix_list is not None,
                self.cidr is not None,
                self.security_group_id is not None,
            ]
        ):
            raise ValueError(
                "A rule is either a CIDR, a security group id or a prefix list!"
            )
        if (
            self.prefix_list is None
            and self.cidr is None
            and self.security_group_id is None
        ):
            raise ValueError(
                "A rule is either a CIDR, a security group id or a prefix list, not"
                " none!"
            )
        return self


class SecurityGroupRules(BaseModel):
    egress: List[SecurityGroupRule] = []
    ingress: List[SecurityGroupRule] = []


class EcsLoadBalancerModel(BaseModel):
    listener_port: int
    target_group_port: int
    subnets: list[str] = []
    load_balancer_name: Optional[str] = None
    arn: Optional[str] = None
    security_group_id: Optional[str] = None
    security_group_rules: Optional[SecurityGroupRules] = None

    @model_validator(mode="after")
    def validate_subnets(self):
        if len(self.subnets) == 0 and self.arn is None:
            raise ValueError(
                "If you do not provide an ARN of a NetworkLoadBalancer, please provide"
                " a subnets list to create one"
            )
        return self


class EcsFileModel(BaseModel):
    metadata: EcsFileMetadataModel
    role: EcsFileRoleModel
    execution_role: EcsFileRoleModel
    task_definition: EcsTaskDefinitionModel
    load_balancer: Optional[EcsLoadBalancerModel] = None
    security_group_id: Optional[str] = None



import datetime
import os
from os.path import dirname, basename
from pathlib import Path
import time
from watchdog.events import FileSystemEvent, FileSystemEventHandler
import tarfile
import socket

from easyecs.helpers.color import Color


def create_tar_for_sync(input, output_dirname, port, volumes_excludes):
    input_path = Path(input)
    with tarfile.open(f"/tmp/{port}.copy.tar.gz", "w:gz") as f:
        if input_path.is_dir():
            for root, dirs, files in os.walk(input_path):
                for file in files:
                    full_path = os.path.join(root, file)
                    rel_path = os.path.relpath(full_path, input_path)
                    archive_name = os.path.join(output_dirname, rel_path)
                    found_exclude = False
                    for volume_exclude in volumes_excludes:
                        if volume_exclude in full_path:
                            found_exclude = True
                            break
                    if not found_exclude:
                        f.add(full_path, arcname=archive_name)
        else:
            filename = basename(input)
            f.add(input, arcname=f"{output_dirname}/{filename}")


def netcat(hostname, port, content, input, output):
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    retry = 0
    MAX_RETRY = 3
    while retry < MAX_RETRY:
        try:
            s.connect((hostname, port))
            s.sendall(content)
            s.shutdown(socket.SHUT_WR)
            data = s.recv(1024)
            print(
                f"\n{Color.GRAY}Synchronized {input} to {output} !{Color.END}", end=""
            )
            if len(data) == 0:
                break
            print("Received:", repr(data))
            s.close()
        except Exception:
            retry += 1
            time.sleep(1)


class SynchronizeEventHandler(FileSystemEventHandler):
    def __init__(self, volume, port, volumes_excludes):
        super().__init__()
        self.volume = volume
        self.port = port
        self.input = volume.split(":")[0]
        self.output = volume.split(":")[1]
        self.input_dirname = dirname(self.input)
        self.output_dirname = dirname(self.output)
        self.last_event = datetime.datetime.now().timestamp()
        self.volumes_excludes = volumes_excludes

    def synchronize(self):
        port = int(self.port)
        create_tar_for_sync(
            self.input, self.output_dirname, port, self.volumes_excludes
        )
        f = open(f"/tmp/{port}.copy.tar.gz", "rb")
        data = f.read()
        netcat("127.0.0.1", port, data, self.input, self.output_dirname)

    def dispatch(self, event: FileSystemEvent):
        delta = datetime.datetime.now().timestamp() - self.last_event
        if delta >= 1:
            event_src = event.src_path
            if event.event_type in ["modified", "created", "moved", "deleted"]:
                if not os.path.isdir(event_src):
                    if os.path.isdir(self.input):
                        if os.path.abspath(self.input) in os.path.abspath(event_src):
                            self.synchronize()
                            super().dispatch(event)
                            self.last_event = datetime.datetime.now().timestamp()
                    else:
                        if os.path.abspath(event_src) in os.path.abspath(self.input):
                            self.synchronize()
                            super().dispatch(event)
                            self.last_event = datetime.datetime.now().timestamp()



from itertools import cycle
from math import floor
from shutil import get_terminal_size
from threading import Thread
from time import sleep

from easyecs.helpers.color import Color


class Loader:
    def __init__(self, desc="Loading...", end="Done!", error="Error!", timeout=0.1):
        """
        A loader-like context manager

        Args:
            desc (str, optional): The loader's description. Defaults to "Loading...".
            end (str, optional): Final print. Defaults to "Done!".
            timeout (float, optional): Sleep time between prints. Defaults to 0.1.
        """
        self.desc = desc
        self.end = end
        self.error = error
        self.timeout = timeout
        self.time = 0
        self.metadata = None

        self._thread = Thread(target=self._animate, daemon=True)
        self.steps = ["⢿", "⣻", "⣽", "⣾", "⣷", "⣯", "⣟", "⡿"]
        self.done = False

    def start(self):
        self._thread.start()
        return self

    def _animate(self):
        for c in cycle(self.steps):
            if self.done:
                break
            if not self.metadata:
                print(
                    f"\r{self.desc} {c} {Color.GRAY}[{floor(self.time)}s]{Color.END}",
                    flush=True,
                    end="",
                )
            else:
                print(f"\r{self.metadata}", flush=True)
                self.metadata = None
            sleep(self.timeout)
            self.time += self.timeout

    def __enter__(self):
        self.start()

    def set_metadata(self, metadata):
        self.metadata = metadata

    def stop(self):
        self.done = True
        cols = get_terminal_size((80, 20)).columns
        print("\r" + " " * cols, end="", flush=True)
        if not self.metadata:
            print(f"\r{self.end} [{floor(self.time)}s]", flush=True)
        else:
            print(f"\r{self.metadata}", flush=True)
            print(f"\r{self.end} [{floor(self.time)}s]")
            self.metadata = None

    def stop_error(self):
        self.done = True
        cols = get_terminal_size((80, 20)).columns
        print("\r" + " " * cols, end="", flush=True)
        if not self.metadata:
            print(f"\r{self.error}", flush=True)
        else:
            print(f"\r{self.metadata}", flush=True)
            print(f"\r{self.error}")
            self.metadata = None

    def __exit__(self, exc_type, exc_value, tb):
        # handle exceptions with those variables ^
        self.stop()



from hashlib import md5
import json
import os
import subprocess
import time
import boto3
import signal
from watchdog.observers import Observer
from easyecs.command.event.synchronize_event_handler import SynchronizeEventHandler
from easyecs.helpers.color import Color
from easyecs.helpers.common import generate_random_port, is_port_in_use
from easyecs.helpers.loader import Loader

from easyecs.helpers.signal import override_sigint

port_forward_pids = []
threads = []
event_handlers = []
popen_procs_port_forward = []
popen_procs_exec_command = []


def create_port_forwards(ecs_manifest, aws_region, aws_account, parsed_containers):
    containers = ecs_manifest.task_definition.containers
    for container in containers:
        container_name = container.name
        container_ports = container.port_forward
        for container_port in container_ports:
            from_port = container_port.split(":")[0]
            to_port = container_port.split(":")[1]
            loader = Loader(
                f"Creating port forward on {container_name} container from"
                f" {from_port} to {to_port}:",
                f"Creating port forward on {container_name} container from"
                f" {from_port} to {to_port}: \u2705",
                f"Creating port forward on {container_name} container from"
                f" {from_port} to {to_port}: \u274c",
                0.05,
            )
            loader.start()
            if not is_port_in_use(int(from_port)):
                port_forward(
                    parsed_containers,
                    container_name,
                    to_port,
                    from_port,
                    aws_region,
                    aws_account,
                )
                loader.stop()
            else:
                loader.stop_error()
                print(f"{Color.RED}Port {from_port} is already in use!{Color.END}")


def port_forward(
    parsed_containers,
    container_name,
    port_number,
    local_port_number,
    aws_region,
    aws_account,
):
    container = parsed_containers.get(container_name, None)
    if container:
        target = container["ssm_target"]
        client = boto3.client("ssm")
        ssm_response = client.start_session(
            Target=target,
            DocumentName="AWS-StartPortForwardingSessionToRemoteHost",
            Parameters={
                "host": ["localhost"],
                "portNumber": [port_number],
                "localPortNumber": [local_port_number],
            },
        )
        # It has to be done like that, in a new session.
        # Otherwise a CTRL+C would kill all port forwards.
        ssm_cmd = generate_ssm_cmd(ssm_response, aws_region, aws_account, target)
        process = subprocess.Popen(
            ssm_cmd,
            start_new_session=True,
            stdin=subprocess.PIPE,
            stdout=subprocess.DEVNULL,
        )
        popen_procs_port_forward.append(process)


def run_sync_thread(parsed_containers, ecs_manifest):
    containers = ecs_manifest.task_definition.containers
    for container in containers:
        if len(container.volumes) > 0:
            observer = Observer()
            container_name = container.name
            for volume in container.volumes:
                md5_volume = md5(volume.encode("utf-8")).hexdigest()
                port = parsed_containers[container_name].get(
                    f"netcat_port_{md5_volume}", None
                )
                _from, _ = volume.split(":")
                event_handler = SynchronizeEventHandler(
                    volume, port, container.volumes_excludes
                )
                event_handlers.append(event_handler)
                observer.schedule(event_handler, _from, recursive=True)
            observer.daemon = True
            observer.start()
            threads.append(observer)


def execute_command(ecs_manifest, parsed_containers, aws_region, aws_account):
    containers = ecs_manifest.task_definition.containers
    catchable_sigs = set(signal.Signals) - {signal.SIGKILL, signal.SIGSTOP}
    ssm_client = boto3.client("ssm")
    found_tty = False
    tty_cmd = ""
    for container in containers:
        command = container.command
        tty = container.tty
        if command and tty:
            tty = container.tty
            container_name = container.name
            container_command = container.command
            target = parsed_containers.get(container_name)["ssm_target"]
            parameters_nc_server = {"command": [container_command]}
            ssm_container = ssm_client.start_session(
                Target=target,
                DocumentName="AWS-StartInteractiveCommand",
                Parameters=parameters_nc_server,
            )
            cmd_container = [
                "session-manager-plugin",
                json.dumps(ssm_container),
                aws_region,
                "StartSession",
                aws_account,
                json.dumps(dict(Target=target)),
                "https://ssm.eu-west-1.amazonaws.com",
            ]
            if tty:
                found_tty = True
                tty_cmd = cmd_container
            else:
                proc_nc_server = subprocess.Popen(
                    cmd_container,
                    stdin=subprocess.PIPE,
                    start_new_session=True,
                )
                popen_procs_exec_command.append(proc_nc_server)
    if found_tty:
        for sig in catchable_sigs:
            signal.signal(sig, override_sigint)
        proc_nc_server = subprocess.Popen(tty_cmd)
        while True:
            if not proc_nc_server.poll():
                proc_nc_server.wait()
                break
            time.sleep(0.1)
    return found_tty


def generate_ssm_cmd(ssm_nc_server, aws_region, aws_account, target):
    return [
        "session-manager-plugin",
        json.dumps(ssm_nc_server),
        aws_region,
        "StartSession",
        aws_account,
        json.dumps(dict(Target=target)),
        f"https://ssm.{aws_region}.amazonaws.com",
    ]


def run_nc_command(
    parsed_containers, aws_region, aws_account, container_name, ecs_manifest
):
    containers = ecs_manifest.task_definition.containers
    for container in containers:
        if len(container.volumes) > 0:
            for volume in container.volumes:
                md5_volume = md5(volume.encode("utf-8")).hexdigest()
                random_port = generate_random_port()
                parsed_containers[container_name][
                    f"netcat_port_{md5_volume}"
                ] = random_port
                port_forward(
                    parsed_containers,
                    container_name,
                    random_port,
                    random_port,
                    aws_region,
                    aws_account,
                )
                client = boto3.client("ssm")
                target = parsed_containers.get(container_name)["ssm_target"]
                command_server = [f"""
                    bash -c '
                        set -x
                        set -u
                        while true
                        do
                            RANDOM_PORT={random_port}
                            RANDOM_NUMBER=$RANDOM
                            nc -v -l ${{RANDOM_PORT}} > /tmp/${{RANDOM_PORT}}.${{RANDOM_NUMBER}}.tar.gz.tmp
                            cp /tmp/${{RANDOM_PORT}}.${{RANDOM_NUMBER}}.tar.gz.tmp /tmp/${{RANDOM_PORT}}.${{RANDOM_NUMBER}}.copy.tar.gz
                            rm /tmp/${{RANDOM_PORT}}.${{RANDOM_NUMBER}}.tar.gz.tmp
                            fc=$(cat /tmp/${{RANDOM_PORT}}.${{RANDOM_NUMBER}}.copy.tar.gz | tar -ztf - | head -c1)
                            if [ $fc = . ]
                            then
                                cat /tmp/${{RANDOM_PORT}}.${{RANDOM_NUMBER}}.copy.tar.gz | tar -xzf -
                            else
                                # cat /tmp/${{RANDOM_PORT}}.${{RANDOM_NUMBER}}.copy.tar.gz | tar -xzf - -C /
                                tar -xzf /tmp/${{RANDOM_PORT}}.${{RANDOM_NUMBER}}.copy.tar.gz -C /
                            fi
                            rm /tmp/${{RANDOM_PORT}}.${{RANDOM_NUMBER}}.copy.tar.gz
                        done'
                    """]  # noqa
                parameters_nc_server = {"command": command_server}
                ssm_nc_server = client.start_session(
                    Target=target,
                    DocumentName="AWS-StartInteractiveCommand",
                    Parameters=parameters_nc_server,
                )
                cmd_nc_server = generate_ssm_cmd(
                    ssm_nc_server, aws_region, aws_account, target
                )
                DEBUG_EASYECS = os.environ.get("DEBUG_EASYECS", None)
                stdout = None if DEBUG_EASYECS else subprocess.DEVNULL
                proc_nc_server = subprocess.Popen(
                    cmd_nc_server,
                    start_new_session=True,
                    stdin=subprocess.PIPE,
                    stdout=stdout,
                )
                popen_procs_port_forward.append(proc_nc_server)


def install_netcat_command(target, aws_region, aws_account) -> None:
    client = boto3.client("ssm")
    commands_server = [["apt update"], ["apt install -y netcat-openbsd"]]
    for command_server in commands_server:
        parameters_nc_server = {"command": command_server}
        ssm_nc_server = client.start_session(
            Target=target,
            DocumentName="AWS-StartInteractiveCommand",
            Parameters=parameters_nc_server,
        )
        cmd_nc_server = [
            "session-manager-plugin",
            json.dumps(ssm_nc_server),
            aws_region,
            "StartSession",
            aws_account,
            json.dumps(dict(Target=target)),
            "https://ssm.eu-west-1.amazonaws.com",
        ]
        DEBUG_EASYECS = os.environ.get("DEBUG_EASYECS", None)
        stdout = None if DEBUG_EASYECS else subprocess.DEVNULL
        subprocess.run(
            cmd_nc_server,
            start_new_session=True,
            stdout=stdout,
        )


def check_nc_command(target, aws_region, aws_account):
    client = boto3.client("ssm")
    command_server = ["which nc"]
    parameters_nc_server = {"command": command_server}
    ssm_nc_server = client.start_session(
        Target=target,
        DocumentName="AWS-StartInteractiveCommand",
        Parameters=parameters_nc_server,
    )
    cmd_nc_server = [
        "session-manager-plugin",
        json.dumps(ssm_nc_server),
        aws_region,
        "StartSession",
        aws_account,
        json.dumps(dict(Target=target)),
        "https://ssm.eu-west-1.amazonaws.com",
    ]
    output = subprocess.check_output(cmd_nc_server, start_new_session=True)
    return "/nc" in output.decode("utf8").split("\n")[2]


def run_nc_commands(
    parsed_containers, aws_region, aws_account, ecs_manifest, auto_install_nc
):
    containers = ecs_manifest.task_definition.containers
    for container in containers:
        if len(container.volumes) > 0:
            container_name = container.name
            parsed_container = parsed_containers.get(container_name)
            ssm_target = parsed_container["ssm_target"]
            loader = Loader(
                f"Running netcat command on container {container_name} for"
                " synchronization:",
                f"Running netcat command on container {container_name} for"
                " synchronization: \u2705",
                f"Running netcat command on container {container_name} for"
                " synchronization: \u274c",
                0.05,
            )
            loader.start()
            has_nc = check_nc_command(ssm_target, aws_region, aws_account)
            if not has_nc and not auto_install_nc:
                loader.stop_error()
                print(
                    f"{Color.YELLOW}In order to use volumes on container"
                    f" {container_name}, you need to install netcat command on the"
                    " container and on the host machine!\nYou can try to install it on"
                    f" the container using --auto-install-nc{Color.END}"
                )
            else:
                if not has_nc and auto_install_nc:
                    install_netcat_command(ssm_target, aws_region, aws_account)
                run_nc_command(
                    parsed_containers,
                    aws_region,
                    aws_account,
                    container_name,
                    ecs_manifest,
                )
                run_sync_thread(parsed_containers, ecs_manifest)
                loader.stop()


def run_force_new_deployment(stack_name):
    cluster_name = f"{stack_name}-cluster"
    service_name = f"{stack_name}-service"
    client = boto3.client("ecs")
    loader = Loader(
        "Force new deployment on service:",
        "Force new deployment on service: \u2705",
        "Force new deployment on service: \u274c",
        0.05,
    )
    loader.start()
    client.update_service(
        cluster=cluster_name,
        service=service_name,
        forceNewDeployment=True,
        deploymentConfiguration={"minimumHealthyPercent": 0},
    )
    waiter = client.get_waiter("services_stable")
    waiter.wait(
        cluster=cluster_name,
        services=[service_name],
        WaiterConfig={"Delay": 2, "MaxAttempts": 1000},
    )
    loader.stop()



from easyecs.cloudformation.client import get_client_cloudformation


def wait_for_stack_update(stack_name: str):
    """
    Waits for the CloudFormation stack to be updated.
    Throws an exception if the update fails.
    """
    client = get_client_cloudformation()
    waiter = client.get_waiter("stack_update_complete")
    waiter.wait(StackName=stack_name)


def wait_for_stack_rollback(stack_name: str):
    """
    Waits for the CloudFormation stack to be rolled back.
    """
    client = get_client_cloudformation()
    waiter = client.get_waiter("stack_rollback_complete")
    waiter.wait(StackName=stack_name)


def wait_for_stack_create(stack_name: str):
    """
    Waits for the CloudFormation stack to be created.
    """
    client = get_client_cloudformation()
    waiter = client.get_waiter("stack_create_complete")
    waiter.wait(StackName=stack_name)


def wait_for_stack_delete(stack_name: str):
    """
    Waits for the CloudFormation stack to be deleted.
    """
    client = get_client_cloudformation()
    waiter = client.get_waiter("stack_delete_complete")
    waiter.wait(StackName=stack_name)



import re
from easyecs.model.ecs import EcsFileSecretModel, EcsFileSecretModelV2


def create_task_definition(
    stack, service_name, task_role, execution_role, log_group, ecs_data, run=False
):
    from aws_cdk.aws_ecs import (
        LogDriver,
    )

    # Create task definition
    task_definition = create_fargate_task_definition(
        stack, service_name, task_role, execution_role, ecs_data
    )

    # Add EFS volumes to task definition
    add_volumes_to_task_definition(task_definition, ecs_data)

    # Configure logging for containers
    log_configuration = LogDriver.aws_logs(stream_prefix="ecs", log_group=log_group)

    # Add containers to task definition
    dict_container_definitions = add_containers_to_task_definition(
        stack, task_definition, ecs_data, log_configuration, run
    )

    # Add dependencies between containers
    add_container_dependencies(dict_container_definitions)

    return task_definition


def create_fargate_task_definition(
    stack, service_name, task_role, execution_role, ecs_data
):
    from aws_cdk.aws_ecs import (
        FargateTaskDefinition,
    )

    """Create a Fargate task definition."""
    task_definition_name = f"{service_name}-task-definition"
    resource_limits = ecs_data.task_definition.resources.limits
    cpu_limit = resource_limits.cpu * 1024
    memory_limit = resource_limits.memory

    return FargateTaskDefinition(
        stack,
        task_definition_name,
        task_role=task_role,
        execution_role=execution_role,
        cpu=cpu_limit,
        memory_limit_mib=memory_limit,
    )


def add_volumes_to_task_definition(task_definition, ecs_data):
    from aws_cdk.aws_ecs import EfsVolumeConfiguration, AuthorizationConfig

    """Add EFS volumes to the task definition."""
    for volume in ecs_data.task_definition.efs_volumes:
        efs_volume_configuration = EfsVolumeConfiguration(
            file_system_id=volume.id,
            authorization_config=AuthorizationConfig(iam="ENABLED"),
            transit_encryption="ENABLED",
        )
        task_definition.add_volume(
            name=volume.name, efs_volume_configuration=efs_volume_configuration
        )


def add_containers_to_task_definition(
    stack, task_definition, ecs_data, log_configuration, run
):
    """Add containers to the task definition."""
    from aws_cdk import aws_ecs

    container_definitions = ecs_data.task_definition.containers
    dict_container_definitions = {}

    for container_definition in container_definitions:
        container_config = extract_container_config(
            stack, container_definition, log_configuration, run
        )
        container = task_definition.add_container(**container_config)
        for port in container_definition.ports:
            container_port = port.split(":")[1]
            host_port = port.split(":")[0]
            container.add_port_mappings(
                aws_ecs.PortMapping(
                    container_port=int(container_port), host_port=int(host_port)
                )
            )
        add_mount_points_to_container(container, container_definition.efs_volumes)
        dict_container_definitions[container_definition.name] = {
            "container": container,
            "container_definition": container_definition,
        }

    return dict_container_definitions


def extract_container_config(stack, container_definition, log_configuration, run):
    from aws_cdk.aws_ecs import (
        ContainerImage,
    )

    """Extract container configuration from its definition."""
    name = container_definition.name
    image = container_definition.image
    user = container_definition.user
    essential = container_definition.essential
    resource_limits = container_definition.resources.limits
    tty = container_definition.tty
    cpu = resource_limits.cpu * 1024
    memory = resource_limits.memory
    entry_point = split_if_str(container_definition.entry_point)
    command = split_if_str(container_definition.command)

    if tty and not run:
        command = ["sleep", "infinity"]

    environment = {}
    if isinstance(container_definition.env, list):
        environment = {
            env_definition.name: env_definition.value
            for env_definition in container_definition.env
            if env_definition.active
        }
    elif isinstance(container_definition.env, dict):
        environment = container_definition.env

    secrets = extract_secrets(stack, container_definition.secrets, name)

    health_check = extract_health_check(container_definition.healthcheck)

    return {
        "id": name,
        "container_name": name,
        "image": ContainerImage.from_registry(image),
        "command": command,
        "logging": log_configuration,
        "environment": environment,
        "secrets": secrets,
        "cpu": cpu,
        "memory_limit_mib": memory,
        "user": user,
        "entry_point": entry_point,
        "health_check": health_check,
        "essential": essential,
    }


def split_if_str(value):
    """Split the value if it's a string."""
    return value.split(" ") if isinstance(value, str) else value


def extract_secrets(stack, secret_definitions, container_name):
    from aws_cdk.aws_secretsmanager import Secret
    from aws_cdk.aws_ecs import (
        Secret as ECSSecret,
    )

    """Extract container secrets from its definition."""
    secrets = {}
    for secret_definition in secret_definitions:
        if isinstance(secret_definition, EcsFileSecretModel):
            secret_name = secret_definition.name
            secret = Secret.from_secret_complete_arn(
                stack, f"{secret_name}_{container_name}", secret_definition.arn
            )
            ecs_secret = ECSSecret.from_secrets_manager(secret, secret_definition.field)
            secrets[secret_name] = ecs_secret
        elif isinstance(secret_definition, EcsFileSecretModelV2):
            arn_fields = list(
                re.finditer(
                    r"^^(?P<secret_complete_arn>arn:aws:secretsmanager:(?P<region_name>[a-z0-9-]+):(?P<account_id>\d{12}):secret:(?P<secret_name>[^:]+))(?::(?P<field>[^:]*))?(?::([^:]*))?(?::([^:]*))?$",  # noqa
                    secret_definition.valueFrom,
                )
            )
            if not arn_fields:
                raise ValueError(f"Invalid ARN format: {secret_definition.valueFrom}")
            secret_complete_arn = arn_fields[0].groupdict()["secret_complete_arn"]
            field = arn_fields[0].groupdict()["field"]
            secret_name = secret_definition.name
            secret = Secret.from_secret_complete_arn(
                stack, f"{secret_name}_{container_name}", secret_complete_arn
            )
            ecs_secret = ECSSecret.from_secrets_manager(secret, field=field)
            secrets[secret_name] = ecs_secret
        else:
            raise Exception("Unsupported secret type")
    return secrets


def extract_health_check(raw_healthcheck):
    from aws_cdk.aws_ecs import (
        HealthCheck,
    )
    from aws_cdk import Duration

    """Extract health check from container definition."""
    if not raw_healthcheck:
        return None

    healthcheck_command = split_if_str(raw_healthcheck.command)

    return HealthCheck(
        command=healthcheck_command,
        interval=Duration.seconds(raw_healthcheck.interval),
        retries=raw_healthcheck.retries,
        start_period=Duration.seconds(raw_healthcheck.start_period),
        timeout=Duration.seconds(raw_healthcheck.timeout),
    )


def add_mount_points_to_container(container, volumes):
    from aws_cdk.aws_ecs import (
        MountPoint,
    )

    """Add mount points to a container."""
    for volume in volumes:
        mount_point = MountPoint(
            container_path=volume.mount_point,
            read_only=False,
            source_volume=volume.name,
        )
        container.add_mount_points(mount_point)


def add_container_dependencies(dict_container_definitions):
    from aws_cdk.aws_ecs import (
        ContainerDependency,
    )

    """Add dependencies between containers."""
    for dict_container in dict_container_definitions.values():
        container = dict_container["container"]
        container_definition = dict_container["container_definition"]

        if container_definition.depends_on:
            for key, value in container_definition.depends_on.items():
                dependency_container = dict_container_definitions[key]["container"]
                container_dependency_condition = map_dependency_condition(
                    value["condition"]
                )
                container_dependency = ContainerDependency(
                    container=dependency_container,
                    condition=container_dependency_condition,
                )
                container.add_container_dependencies(container_dependency)


def map_dependency_condition(condition_str):
    from aws_cdk.aws_ecs import ContainerDependencyCondition

    """Map a dependency condition to the appropriate AWS CDK condition."""
    if condition_str == "service_completed_successfully":
        return ContainerDependencyCondition.COMPLETE
    elif condition_str == "service_healthy":
        return ContainerDependencyCondition.HEALTHY
    else:
        raise ValueError(f"Unrecognized dependency condition: {condition_str}")
